#include <stdio.h>
#include "cuda.h"
#include "mycudautils.h"

/*

 http://www.willamette.edu/~sekino/fractal/#I1

 Gatekeepers in Figure 1 above, for example, is a Julia pattern given by

 z(n+1)= z(n)^3 + z(n) + p,   p = (0.185, 0.00007666),

on the rectangular region with xmin = -1.1, xmax = 1.1, ymin = -0.7, and ymax = 0.7. 

*/

extern "C"
{
#include "png_util.h"
}

#define Nloops 800

/* cuda julia kernel */
__global__ void julia2D(float remin, float immin, float dre, float dim, float pre, float pim, float *d_count){
  
  /* thread coordinates */
  const int re_id = blockIdx.x*blockDim.x + threadIdx.x;
  const int im_id = blockIdx.y*blockDim.y + threadIdx.y;

  /* value of the complex variable c */
  float zre = remin + dre*re_id;
  float zim = immin + dim*im_id;

  const float escape = 4;
  
  /* iterate */
  int loop=0, count = 0;;
  for(loop=0;loop<Nloops;++loop){
    /* update z */
    const float z2re = zre*zre - zim*zim;
    const float z2im = 2.f*zre*zim;
    const float z3re = zre*z2re - zim*z2im;
    const float z3im = zre*z2im + zim*z2re;

    zre = z3re + zre + pre;
    zim = z3im + zim + pim;
    
    /* escape test */
#if 0
    const int test =  (zre*zre+zim*zim>escape);
    if(test)
      break;
#else
    const int test =  (zre*zre+zim*zim<escape);
    count += test;
#endif
  }
 
  /* output result */
  const int Nre = blockDim.x*gridDim.x;
#if 0 
  d_count[re_id + im_id*Nre] = sqrtf( (float)loop);
#else
  d_count[re_id + im_id*Nre] = sqrtf( (float)count);
#endif
}

int main(){

  int Bre = 256;
  int Bim = 256;
  int Tre = 16;
  int Tim = 16;
  
  /* number of pixels in real and imaginary directions */
  int Nre = Bre*Tre;
  int Nim = Bim*Tim;

  cudaSetDevice(0);

  /* specify box in complex plane */
  const float centRe = 0, centIm= 0;
  const float diamRe  = 2.2;
  const float diamIm  = 2.2;

  const float remin = centRe-0.5*diamRe;
  const float remax = centRe+0.5*diamRe;
  const float immin = centIm-0.5*diamIm;
  const float immax = centIm+0.5*diamIm;

  const float dre = (remax-remin)/(Nre-1.f);
  const float dim = (immax-immin)/(Nim-1.f);

  const float pre = 0.185, pim = 0.00007666;

  /* dimension of block grid */
  dim3 dimBlockGrid(Bre,Bim);

  /* dimension of thread block */
  dim3 dimThreadBlock(Tre,Tim);

  /* host arrays */
  float *h_count = (float*) calloc(Bre*Bim*Tre*Tim, sizeof(float));

  /* device arrays */
  float *d_count;

  /* allocate device arrays */
  int N = Bre*Tre*Bim*Tim;
  int sz = N*sizeof(float);
  myCudaMalloc( (void**) &d_count, sz, "allocate d_count");

  /* start timer */
  cudaEvent_t tic, toc;
  StartKernelTiming(tic, toc, 0);

  /* request to invoke kernel */
  julia2D <<< dimBlockGrid, dimThreadBlock >>> ( immin, remin, dim, dre, pre, pim, d_count);

  /* stop timer */
  float elapsed_time = 0.f;
  StopKernelTiming(tic,toc, 0, &elapsed_time);

  /* convert from miliseconds to seconds */
  elapsed_time /= 1000.f;

  /* output elapsed time */
  printf("elapsed time for kernel: %f\n", elapsed_time);
  printf("performance: %g GFLOPS\n", (10.*Nre*Nim*Nloops)/(1.e9*elapsed_time));

  myCudaDeviceToHost( h_count, d_count, sz, "copy result to host");

  FILE *png = fopen("juliaGPU.png", "w");
  write_hot_png(png, Nre, Nim, h_count, 0, sqrtf((float)Nloops));
  fclose(png);

  return 1;
}
