#include <stdio.h>
#include "cuda.h"
#include "mycudautils.h"

extern "C"
{
#include "png_util.h"
}

/* cuda mandelbrot kernel */
__global__ void mandelbrot2D(float remin, float immin, float dre, float dim, float *d_count){
  
  /* thread coordinates */
  const int re_id = blockIdx.x*blockDim.x + threadIdx.x;
  const int im_id = blockIdx.y*blockDim.y + threadIdx.y;

  /* value of the complex variable c */
  const float cre = remin + dre*re_id;
  const float cim = immin + dim*im_id;

  const float escape = 4;
  float  zre = 0.f, zim = 0.f;
  
  /* iterate */
  const int Nloops = 6400;
  int loop=0, test1, test2;
  for(loop=0;loop<Nloops;++loop){
    /* update z */
    float new_zre =    zre*zre - zim*zim + cre;
    float new_zim =          2.f*zre*zim + cim;
    
    zre = new_zre;
    zim = new_zim;
    
    /* escape test */
    test1 =  (zre*zre+zim*zim>escape);

    /* update z */
    new_zre =    zre*zre - zim*zim + cre;
    new_zim =          2.f*zre*zim + cim;
    
    zre = new_zre;
    zim = new_zim;
    
    /* escape test */
    test2 =  (zre*zre+zim*zim>escape);
    if(test1+test2)
      break;

  }
 
  /* output result */
  const int Nre = blockDim.x*gridDim.x;
  d_count[re_id + im_id*Nre] = sqrtf( (float)(loop+test2-test1) );
}

main(){

  int Bre = 1200;
  int Bim = 1200;
  int Tre = 16;
  int Tim = 16;
  
  /* number of pixels in real and imaginary directions */
  int Nre = Bre*Tre;
  int Nim = Bim*Tim;

  /* specify box in complex plane */
  const float centRe = -.759856, centIm= .125547;
  const float diam  = 0.151579;
  const float remin = centRe-0.5*diam;
  const float remax = centRe+0.5*diam;
  const float immin = centIm-0.5*diam;
  const float immax = centIm+0.5*diam;

  const float dre = (remax-remin)/(Nre-1.f);
  const float dim = (immax-immin)/(Nim-1.f);

  /* dimension of block grid */
  dim3 dimBlockGrid(Bre,Bim);

  /* dimension of thread block */
  dim3 dimThreadBlock(Tre,Tim);

  /* host arrays */
  float *h_count = (float*) calloc(Bre*Bim*Tre*Tim, sizeof(float));

  /* device arrays */
  float *d_count;

  /* allocate device arrays */
  int N = Bre*Tre*Bim*Tim;
  int sz = N*sizeof(float);
  myCudaMalloc( (void**) &d_count, sz, "allocate d_count");

  /* start timer */
  cudaEvent_t tic, toc;
  StartKernelTiming(tic, toc, 0);

  /* request to invoke kernel */
  mandelbrot2D <<< dimBlockGrid, dimThreadBlock >>> ( remin, immin, dre, dim, d_count);

  /* stop timer */
  float elapsed_time = 0.f;
  StopKernelTiming(tic,toc, 0, &elapsed_time);

  /* convert from miliseconds to seconds */
  elapsed_time /= 1000.f;

  /* output elapsed time */
  printf("elapsed time for kernel: %f\n", elapsed_time);
  //  printf("performance: %g GFLOPS\n", (10.*Nre*Nim*Nloops)/(1.e9*elapsed_time));

  myCudaDeviceToHost( h_count, d_count, sz, "copy result to host");

  FILE *png = fopen("mandelbrotGPU.png", "w");
  write_hot_png(png, Nre, Nim, h_count, 0, 80);
  fclose(png);
}
