
#include "mycudautils.h"

#define NTHREADS 256
#define STRIDE (4*NTHREADS)

__global__ void reduce_smem(int N, float *a, float *b){
  
  __shared__ float sa[NTHREADS];

  const int tid = threadIdx.x; 
  const int nid = tid + (STRIDE)*blockIdx.x;
  sa[tid] = 0.f;
  if(nid+3*NTHREADS<N)
    sa[tid] = a[nid] + a[nid+NTHREADS] + a[nid+2*NTHREADS] + a[nid+3*NTHREADS];

  __syncthreads();  if(tid<128) sa[tid] += sa[tid +128];
  __syncthreads();  if(tid<64) sa[tid] += sa[tid + 64];
  __syncthreads();  if(tid<32) sa[tid] += sa[tid + 32];
  __syncthreads();  if(tid<16) sa[tid] += sa[tid + 16];
  __syncthreads();  if(tid<8 ) sa[tid] += sa[tid + 8];
  __syncthreads();  if(tid<4 ) sa[tid] += sa[tid + 4];
  //  __syncthreads();  if(tid<2 ) sa[tid] += sa[tid + 2];
  __syncthreads(); 

  if(tid==0 && nid<N) 
    b[blockIdx.x] = sa[0]+sa[1]+sa[2]+sa[3];
}

__global__ void reduce_smem_v6(int N, float *a, float *b){
  
  __shared__ float s_a[NTHREADS];
  unsigned int tid = threadIdx.x; 
  unsigned int nid = blockIdx.x*(NTHREADS*2) + tid; 
  unsigned int gridSize = NTHREADS*2*gridDim.x; 

  s_a[tid] = 0;

  while (nid < N){
    s_a[tid] += a[nid] + a[nid+NTHREADS]; 
    nid += gridSize; 
  } 
  __syncthreads();

  if (NTHREADS >= 512) { if (tid < 256) { s_a[tid] += s_a[tid + 256]; } __syncthreads(); } 
  if (NTHREADS >= 256) { if (tid < 128) { s_a[tid] += s_a[tid + 128]; } __syncthreads(); } 
  if (NTHREADS >= 128) { if (tid <  64) { s_a[tid] += s_a[tid +  64]; } __syncthreads(); } 

  if (tid < 32) { 
    if (NTHREADS >= 64) s_a[tid] += s_a[tid+32];
    if (NTHREADS >= 32) s_a[tid] += s_a[tid+16];
    if (NTHREADS >= 16) s_a[tid] += s_a[tid+ 8];
    if (NTHREADS >=  8) s_a[tid] += s_a[tid+ 4];
    if (NTHREADS >=  4) s_a[tid] += s_a[tid+ 2];
    if (NTHREADS >=  2) s_a[tid] += s_a[tid+ 1];
  }

  if (tid == 0) b[blockIdx.x] = s_a[0];
}

main(){

 int N = 1 << 24;

  int n;

  // fill host array
  float *h_a = (float*) calloc(N, sizeof(float));
  for(n=0;n<N;++n){
    h_a[n] = 1;
  }

  float *d_a, *d_b;

  cudaSetDevice(0);

  // allocate device array and copy from host to device 
  myCudaMalloc( (void**) &d_a, N*sizeof(float), "allocate d_a");
  myCudaMalloc( (void**) &d_b, N*sizeof(float), "allocate d_b");

  myCudaHostToDevice(d_a, h_a, N*sizeof(float), "copy h_a to d_a");

  /* start timer */
  cudaEvent_t tic, toc;
  StartKernelTiming(tic, toc, 0);

  int Nblocks = (N+STRIDE-1)/(STRIDE);
  
  dim3 dimThreadBlock(NTHREADS);
  dim3 dimGridBlock(Nblocks);

  reduce_smem <<< dimGridBlock, dimThreadBlock >>> (N, d_a, d_b);
  //  reduce_smem_v6 <<< dimGridBlock, dimThreadBlock >>> (N, d_a, d_b);

  myCudaDeviceToHost(h_a, d_b, Nblocks*sizeof(float), "d_b to h_a");

  float res = 0.f;
  for(n=0;n<Nblocks;++n)
    res += h_a[n];

  /* stop timer */
  float elapsed_time = 0.f;
  StopKernelTiming(tic,toc, 0, &elapsed_time);

  /* convert from miliseconds to seconds */
  elapsed_time /= 1000.f;

  /* output elapsed time */
  printf("elapsed time for kernel: %f\n", elapsed_time);
  printf("estimated gflops = %f\n", N/(1e9*elapsed_time));
  printf("estimated bwidth = %f\n", (4*N)/(1e9*elapsed_time));

  printf("reduce(a) = %f exact = %f\n", res, 1.*N);

  cudaFree(d_a);
  free(h_a);
}
