#include <stdio.h>
#include "cuda.h"

// PURPOSE: demonstrate simple CUDA program to initialize data arrays on 
//          DEVICE, add them together, and then return result back to HOST

__global__ void VectorAddKernel(int N, float *a, float *b, float *aPb){

  // Thread index in 1D thread-block
  const int tx = threadIdx.x;

  // Number of threads in 1D thread-block
  const int dimx = blockDim.x;
  
  // Index of thread-block in grid
  const int bx = blockIdx.x;

  // Convert thread and thread-block indices into array index 
  const int n  = tx + bx*dimx;
	
  // If index is in [0,N-1] add entries
  if(n<N)
    aPb[n] = a[n]+b[n];
}

main(){

  // size of arrays for this DEMO
  int N = 512;

  // HOST arrays 
  float *h_a   = (float*) calloc(N, sizeof(float));
  float *h_b   = (float*) calloc(N, sizeof(float));

  // Set data on HOST
  for(int n=0;n<N;++n){
    h_a[n] = n;
    h_b[n] = 1.f-n;
  }
  
  // Allocate DEVICE arrays
  float *d_a, *d_b, *d_aPb;
  cudaMalloc((void**) &d_a,   N*sizeof(float));
  cudaMalloc((void**) &d_b,   N*sizeof(float));
  cudaMalloc((void**) &d_aPb, N*sizeof(float));

  // Copy data from HOST arrays to DEVICE arrays (read from right to left)
  cudaMemcpy(d_a, h_a, N*sizeof(float), cudaMemcpyHostToDevice);
  cudaMemcpy(d_b, h_b, N*sizeof(float), cudaMemcpyHostToDevice);

  // Invoke kernel with specified thread array
  dim3 ThreadsPerBlock(512);            // 512 threads per thread-block
  dim3 BlocksPerGrid( (N+512-1)/512 );  // Enough thread-blocks to cover N
  VectorAddKernel <<< BlocksPerGrid, ThreadsPerBlock >>> (N, d_a, d_b, d_aPb);

  // Allocate storage on HOST for result
  float *h_aPb = (float*) calloc(N, sizeof(float));

  // Transfer result from DEVICE to HOST
  cudaMemcpy(h_aPb, d_aPb, N*sizeof(float), cudaMemcpyDeviceToHost);
  
  // Print out result
  for(int n=0;n<N;++n)
    printf("h_aPb[%d] = %f\n", n, h_aPb[n]);

}
