
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "cuda.h"

#include "mycudautils.h"
extern "C"
{
#include "png_util.h"
}

#define PAD 4
#define NT 12

__global__ void update(const int padN, const float lambda, float *d_p, float *d_pold){

  __shared__ float s_p[NT+2*PAD][NT+2*PAD];

  const int bx = blockIdx.x;
  const int by = blockIdx.y;

  const int tx = threadIdx.x;
  const int ty = threadIdx.y;

  const int idx = tx + bx*NT;
  const int idy = ty + by*NT;

  const int n = idx +  padN*idy;
  s_p[ty][tx] = d_p[n];

  __syncthreads();

  if( (tx<PAD) || (tx>=NT+PAD) || (ty<PAD) || (ty>=NT+PAD)) return;

  const float a0 = -1./560;
  const float a1 =  8./315;
  const float a2 = -1./5;
  const float a3 =  8./5;
  const float a4 = -205./72;

  float lap_p = 0.f; 

  lap_p += a0*s_p[ty][tx+4];
  lap_p += a1*s_p[ty][tx+3];
  lap_p += a2*s_p[ty][tx+2];
  lap_p += a3*s_p[ty][tx+1];
  lap_p += a4*s_p[ty][tx+0];
  lap_p += a3*s_p[ty][tx-1];
  lap_p += a2*s_p[ty][tx-2];
  lap_p += a1*s_p[ty][tx-3];
  lap_p += a0*s_p[ty][tx-4];

  lap_p += a0*s_p[ty+4][tx];
  lap_p += a1*s_p[ty+3][tx];
  lap_p += a2*s_p[ty+2][tx];
  lap_p += a3*s_p[ty+1][tx];
  lap_p += a4*s_p[ty+0][tx];
  lap_p += a3*s_p[ty-1][tx];
  lap_p += a2*s_p[ty-2][tx];
  lap_p += a1*s_p[ty-3][tx];
  lap_p += a0*s_p[ty-4][tx];

  const float pnew = 2.f*s_p[ty][tx] -d_pold[n] + lambda*lap_p;

  // save result
  d_pold[n] = pnew;
}

main(){

  int i,j;

  // 10x10 block array
  int Bx = 80, By = 80;

  // NTxNT thread array per block
  int Tx = NT, Ty = NT;

  // FDTD grid 
  int Nx = Bx*Tx, Ny = By*Ty;

  // HOST side initialization
  int N = (Nx+2*PAD)*(Ny+2*PAD);
  float *h_p    = (float*) calloc(N, sizeof(float));
  float *h_pold = (float*) calloc(N, sizeof(float));

  float xmin = -1, xmax = 1, ymin = -1, ymax = 1;
  float dx = (xmax-xmin)/(Nx-1);
  float dy = (ymax-ymin)/(Ny-1);
  for(i=0;i<Nx;++i){
    for(j=0;j<Ny;++j){
      float x = xmin + i*dx;
      float y = ymin + j*dy;
      int idx = i+PAD;
      int idy = j+PAD;
      h_p[idx+idy*(Nx+2*PAD)]    = exp(-40.*(x*x+y*y));
      h_pold[idx+idy*(Nx+2*PAD)] = exp(-40.*(x*x+y*y));
    }
  }

  // define thread dimensions
  dim3 dimGrid(Bx,By,1);
  dim3 dimBlock(Tx+2*PAD,Ty+2*PAD,1);

  cudaSetDevice(0);

  float *d_p, *d_pold;
  myCudaMalloc((void**) &d_p, N*sizeof(float), "allocate d_p");
  myCudaMalloc((void**) &d_pold, N*sizeof(float), "allocate d_pold");
  
  myCudaHostToDevice(d_p,    h_p,    N*sizeof(float), "copy h_p to d_p");
  myCudaHostToDevice(d_pold, h_pold, N*sizeof(float), "copy h_pold to d_pold");

  /* time step info */
  float T = 4;
  float cfl = .05;
  float dt = cfl*dx;
  int Nsteps = (int)ceil(T/dt), tstep;
  dt = T/Nsteps;

  float lambda = dt*dt/(dx*dx);
  for(tstep=0;tstep<Nsteps;tstep+=2){

    cudaEvent_t tic, toc;
    StartKernelTiming(tic, toc, 0);	

    // flip flop storage
    update <<< dimGrid, dimBlock >>> (Nx+2*PAD, lambda, d_p, d_pold);
    update <<< dimGrid, dimBlock >>> (Nx+2*PAD, lambda, d_pold, d_p);

    /* stop timer */
    float elapsed_time = 0.f;
    StopKernelTiming(tic,toc, 0, &elapsed_time);
    
    /* convert from miliseconds to seconds */
    elapsed_time /= 1000.f;
    
    // peak: 177 GB/s *28/ 12B per node
    int iostep = 1000;
    if(!(tstep%iostep)){

      /* output elapsed time */
      printf("elapsed time for kernel: %g\n", elapsed_time);
      printf("estimated perf for kernel: %g\n", (2*Nx*Ny*36.)/(1e9*elapsed_time));

      char name[BUFSIZ];
      FILE *png;

      sprintf(name, "data_%05d.png", tstep/iostep);

      png = fopen(name, "w");

      myCudaDeviceToHost(h_p, d_p, N*sizeof(float), "copy d_p to h_p");

      write_hot_png(png, Nx+2*PAD, Ny+2*PAD, h_p, -.2f, .2f);

      fclose(png);
    }

  }

}
